//---------------------------------------------------------------------------
/*
	Packet Radio Telnet
    Version 1.0

	Copyright (C) 1997~1999 Marco Savegnago All rights reserved.	
	
	This software and all its components may be freely distributed and used, except
	in a business, commercial, government, or institutional environment.
	It may not under any circumstances be sold by anyone other than the author.
	It may be distributed by a commercial company as long as it is
	for no cost.
	No resposability can be charged to the author if the program running hang the		
	computer with a consequent loss of data.
	If you find it useful a free contribution if always accepted!    
*/
#ifndef __QueueH
#define __QueueH

template <typename T>
class TQueue
{
public:
    class   TXQueue
    {
    public:

        enum    xtype
        {
            Empty,
            Full
        };

        xtype x;

        public:
            TXQueue(xtype X)
            :
                x(X)
                {
                }
    };

	TQueue(unsigned long dwlen=2048);           //	ctor
	virtual ~TQueue();				            //	dtor

	bool 	Put(T b);		                    //	Scrive un byte nella coda
	bool 	Put(const T *block, unsigned long dwlen);	//	Scrive un buffer nella coda

	T       Get();												//	Legge un byte dalla coda
	bool 	Get(T *pbyBlock, unsigned long dwMaxLen, unsigned long& rdwlen);	//	Legge un buffer dalla coda
	bool 	Get(T *pbyBlock, unsigned long dwMaxLen, unsigned long& rdwLen, T byDelimiter, bool bIncludeDelimiter=true);
	T 	    Peek();												//	Legge il primo carattere della coda

	unsigned long	Avail() const;			//	Ritorna il numero di byte occupati nella coda
	unsigned long	Free()  const;			//	Ritorna il numero di byte disponibile nella coda

	bool 	IsAvail();				//	Ritorna vero se ci sono byte disponibili nella coda
	bool	IsEmpty();				//	Ritorna vero se non c'e' spazio
	bool	IsFull();				//	Ritorna vero se la coda e' piena

protected:

	T 		    g_next();		//	Funzione che esegue il prelievo in coda del buffer di un item
	void 		p_next(T);	    //	Funzione che esegue l'inserimento in testa del buffer di un item

	unsigned long		dwbufferLen;	//	Dimensione del buffer
	T*		    pbybuffer;		//	Puntatore al buffer

	unsigned long 	    dwstartBuffer;	//	Indice dal quale vengono prelevati i buffer items
	unsigned long		dwendBuffer;	//	Indice dal quale vengono inseriti i buffer item
};

//	ctor
template<typename T>
TQueue<T>::TQueue(unsigned long dwLen)
:
	dwstartBuffer(0),             	    //	Indice di inzio del buffer
	dwendBuffer(0),						//	Indice di fine del buffer
    dwbufferLen(dwLen)
{
	//	Alloca lo spazio per il buffer e assegno la dimensione di lavoro

	pbybuffer 		= new T[dwbufferLen];
}

//	dtor
template<typename T>
TQueue<T>::~TQueue()
{
	//	Libera lo spazio allocato per il buffer

	delete [] pbybuffer;
}

//	Inserisce un byte nella coda

template<typename T>
bool TQueue<T>::Put(T b)
{
	p_next(b);

	return true;
}

//	Inserisce un blocco nella coda

template<typename T>
bool TQueue<T>::Put(const T *block, unsigned long dwLen)
{
	while(dwLen--)
		p_next(*(block++));

	return true;
}

//	Preleva un byte dalla coda

template<typename T>
T TQueue<T>::Get()
{
	return	g_next();
}

//	Preleva un buffer dalla coda

template<typename T>
bool TQueue<T>::Get(T *pbyBlock, unsigned long dwMaxLen, unsigned long& rdwLen)
{
	bool	brc	=	false;

   //	C'e' qualcosa ?

	if((rdwLen	=	Avail()) > 0)
   {
   	    //	Se c'e' piu' roba di quanto mi posso permettere allora diminuisco
        //	la richiesta e lascio il segnale che non ho finito

   	    if(rdwLen>dwMaxLen)
            rdwLen	=	dwMaxLen;
        else
      	    brc	=	true;

		dwMaxLen	=	rdwLen;

		while(dwMaxLen--)
			*(pbyBlock++)=g_next();
    }

	return brc;
}

//	Preleva un buffer dalla coda
template<typename T>
bool TQueue<T>::Get(T *pbyBlock, unsigned long dwMaxLen, unsigned long& rdwLen, T byDelimiter, bool bIncludeDelimiter)
{
	bool	brc	=	false;

   //	C'e' qualcosa ?

	if(Avail() > 0)
    {
        rdwLen	=	0;

		while(dwMaxLen--)
        {
            rdwLen++;

            if((*pbyBlock=g_next(dwTimeout)) == byDelimiter)
            {
                if(!bIncludeDelimiter)
            	    *pbyBlock=NULL;
                else
            	    rdwLen++;

   			    brc		=	true;
                break;
            }

            rdwLen++;
            pbyBlock++;
        }
    }
	return brc;
}


//	Ritorna vero se nella coda c'e' qualcosa di disponibile

template<typename T>
bool TQueue<T>::IsAvail()
{
	return bool(Avail()>0);
}

//	Ritorna il numero di byte occupati nella coda
template<typename T>
unsigned long 	TQueue<T>::Avail() const
{
	return (dwendBuffer >= dwstartBuffer) ?
				(dwendBuffer - dwstartBuffer) :
					((dwbufferLen - dwstartBuffer) + dwendBuffer);
}
//	Ritorna il numero di byte disponibile nella coda
template<typename T>
unsigned long 	TQueue<T>::Free() const
{
	return	(dwbufferLen - Avail());
}

//	Preleva senza rimuovere un byte dalla coda
template<typename T>
T TQueue<T>::Peek()
{
    return *(pbybuffer+dwstartBuffer);
}

//	Get the next byte in the buffer
template<typename T>
T TQueue<T>::g_next()
{
    if(!IsAvail())
        throw(TXQueue(TXQueue::Empty);

    byte b=*(pbybuffer+dwstartBuffer++);

    if( dwstartBuffer >= dwbufferLen )
        dwstartBuffer=0;

    return b;
}

//	Put the next byte in the buffer
template<typename T>
void TQueue<T>::p_next(T b)
{
    if(IsFull())
        throw TXQueue(TXQueue::Full);

    *(pbybuffer+dwendBuffer++)=b;

    if( dwendBuffer >= dwbufferLen )
        dwendBuffer=0;
}

//	Ritorna vero se non c'e' spazio

template<typename T>
bool	TQueue<T>::IsEmpty()
{
	return	bool(Avail() == 0);
}

//	Ritorna vero se la coda e' piena

template<typename T>
bool	TQueue<T>::IsFull()
{
	return	bool((dwbufferLen - Avail()) == 0);
}


//---------------------------------------------------------------------------
#endif  //  #define __QueueH
